@@ -5,14 +5,14 @@ Tests for numerical integration.
 from __future__ import division, print_function, absolute_import
 
 import numpy as np
-from numpy import arange, zeros, array, dot, sqrt, cos, sin, eye, pi, exp, \
-                  allclose
+from numpy import (arange, zeros, array, dot, sqrt, cos, sin, eye, pi, exp,
+                   allclose)
 
-from scipy.lib.six import xrange
+from scipy._lib.six import xrange
 
-from numpy.testing import assert_, TestCase, run_module_suite, \
-        assert_array_almost_equal, assert_raises, assert_allclose, \
-        assert_array_equal, assert_equal
+from numpy.testing import (
+    assert_, TestCase, run_module_suite, assert_array_almost_equal,
+    assert_raises, assert_allclose, assert_array_equal, assert_equal)
 from scipy.integrate import odeint, ode, complex_ode
 
 #------------------------------------------------------------------------------
@@ -35,8 +35,10 @@ class TestOdeint(TestCase):
             self._do_problem(problem)
 
 
-class TestOde(TestCase):
-    # Check integrate.ode
+class TestODEClass(TestCase):
+
+    ode_class = None   # Set in subclass.
+
     def _do_problem(self, problem, integrator, method='adams'):
 
         # ode has callback arguments in different order than odeint
@@ -45,17 +47,30 @@ class TestOde(TestCase):
         if hasattr(problem, 'jac'):
             jac = lambda t, z: problem.jac(z, t)
 
-        ig = ode(f, jac)
+        integrator_params = {}
+        if problem.lband is not None or problem.uband is not None:
+            integrator_params['uband'] = problem.uband
+            integrator_params['lband'] = problem.lband
+
+        ig = self.ode_class(f, jac)
         ig.set_integrator(integrator,
                           atol=problem.atol/10,
                           rtol=problem.rtol/10,
-                          method=method)
+                          method=method,
+                          **integrator_params)
+
         ig.set_initial_value(problem.z0, t=0.0)
         z = ig.integrate(problem.stop_t)
 
+        assert_array_equal(z, ig.y)
         assert_(ig.successful(), (problem, method))
         assert_(problem.verify(array([z]), problem.stop_t), (problem, method))
 
+
+class TestOde(TestODEClass):
+
+    ode_class = ode
+
     def test_vode(self):
         # Check the vode solver
         for problem_cls in PROBLEMS:
@@ -156,27 +171,9 @@ class TestOde(TestCase):
                 assert_allclose(r2.y, 0.2)
 
 
-class TestComplexOde(TestCase):
-    # Check integrate.complex_ode
-    def _do_problem(self, problem, integrator, method='adams'):
+class TestComplexOde(TestODEClass):
 
-        # ode has callback arguments in different order than odeint
-        f = lambda t, z: problem.f(z, t)
-        jac = None
-        if hasattr(problem, 'jac'):
-            jac = lambda t, z: problem.jac(z, t)
-        ig = complex_ode(f, jac)
-        ig.set_integrator(integrator,
-                          atol=problem.atol/10,
-                          rtol=problem.rtol/10,
-                          method=method)
-        ig.set_initial_value(problem.z0, t=0.0)
-        z = ig.integrate(problem.stop_t)
-        z2 = ig.y
-
-        assert_array_equal(z, z2)
-        assert_(ig.successful(), (problem, method))
-        assert_(problem.verify(array([z]), problem.stop_t), (problem, method))
+    ode_class = complex_ode
 
     def test_vode(self):
         # Check the vode solver
@@ -352,6 +349,9 @@ class ODE:
     stop_t = 1
     z0 = []
 
+    lband = None
+    uband = None
+
     atol = 1e-6
     rtol = 1e-5
 
@@ -370,21 +370,21 @@ class SimpleOscillator(ODE):
     m = 1.0
 
     def f(self, z, t):
-        tmp = zeros((2,2), float)
-        tmp[0,1] = 1.0
-        tmp[1,0] = -self.k / self.m
+        tmp = zeros((2, 2), float)
+        tmp[0, 1] = 1.0
+        tmp[1, 0] = -self.k / self.m
         return dot(tmp, z)
 
     def verify(self, zs, t):
         omega = sqrt(self.k / self.m)
-        u = self.z0[0]*cos(omega*t)+self.z0[1]*sin(omega*t)/omega
-        return allclose(u, zs[:,0], atol=self.atol, rtol=self.rtol)
+        u = self.z0[0]*cos(omega*t) + self.z0[1]*sin(omega*t)/omega
+        return allclose(u, zs[:, 0], atol=self.atol, rtol=self.rtol)
 
 
 class ComplexExp(ODE):
     r"""The equation :lm:`\dot u = i u`"""
     stop_t = 1.23*pi
-    z0 = exp([1j,2j,3j,4j,5j])
+    z0 = exp([1j, 2j, 3j, 4j, 5j])
     cmplx = True
 
     def f(self, z, t):
@@ -409,9 +409,73 @@ class Pi(ODE):
 
     def verify(self, zs, t):
         u = -2j * np.arctan(10)
-        return allclose(u, zs[-1,:], atol=self.atol, rtol=self.rtol)
+        return allclose(u, zs[-1, :], atol=self.atol, rtol=self.rtol)
 
-PROBLEMS = [SimpleOscillator, ComplexExp, Pi]
+
+class CoupledDecay(ODE):
+    r"""
+    3 coupled decays suited for banded treatment
+    (banded mode makes it necessary when N>>3)
+    """
+
+    stiff = True
+    stop_t = 0.5
+    z0 = [5.0, 7.0, 13.0]
+    lband = 1
+    uband = 0
+
+    lmbd = [0.17, 0.23, 0.29]  # fictious decay constants
+
+    def f(self, z, t):
+        lmbd = self.lmbd
+        return np.array([-lmbd[0]*z[0],
+                         -lmbd[1]*z[1] + lmbd[0]*z[0],
+                         -lmbd[2]*z[2] + lmbd[1]*z[1]])
+
+    def jac(self, z, t):
+        # The full Jacobian is
+        #
+        #    [-lmbd[0]      0         0   ]
+        #    [ lmbd[0]  -lmbd[1]      0   ]
+        #    [    0      lmbd[1]  -lmbd[2]]
+        #
+        # The lower and upper bandwidths are lband=1 and uband=0, resp.
+        # The representation of this array in packed format is
+        #
+        #    [-lmbd[0]  -lmbd[1]  -lmbd[2]]
+        #    [ lmbd[0]   lmbd[1]      0   ]
+
+        lmbd = self.lmbd
+        j = np.zeros((self.lband + self.uband + 1, 3), order='F')
+
+        def set_j(ri, ci, val):
+            j[self.uband + ri - ci, ci] = val
+        set_j(0, 0, -lmbd[0])
+        set_j(1, 0, lmbd[0])
+        set_j(1, 1, -lmbd[1])
+        set_j(2, 1, lmbd[1])
+        set_j(2, 2, -lmbd[2])
+        return j
+
+    def verify(self, zs, t):
+        # Formulae derived by hand
+        lmbd = np.array(self.lmbd)
+        d10 = lmbd[1] - lmbd[0]
+        d21 = lmbd[2] - lmbd[1]
+        d20 = lmbd[2] - lmbd[0]
+        e0 = np.exp(-lmbd[0] * t)
+        e1 = np.exp(-lmbd[1] * t)
+        e2 = np.exp(-lmbd[2] * t)
+        u = np.vstack((
+            self.z0[0] * e0,
+            self.z0[1] * e1 + self.z0[0] * lmbd[0] / d10 * (e0 - e1),
+            self.z0[2] * e2 + self.z0[1] * lmbd[1] / d21 * (e1 - e2) +
+            lmbd[1] * lmbd[0] * self.z0[0] / d10 *
+            (1 / d20 * (e0 - e2) - 1 / d21 * (e1 - e2)))).transpose()
+        return allclose(u, zs, atol=self.atol, rtol=self.rtol)
+
+
+PROBLEMS = [SimpleOscillator, ComplexExp, Pi, CoupledDecay]
 
 #------------------------------------------------------------------------------
 
@@ -543,6 +607,15 @@ class LSODACheckParameterUse(ODECheckParameterUse, TestCase):
     solver_uses_jac = True
 
 
+def test_odeint_trivial_time():
+    # Test that odeint succeeds when given a single time point
+    # and full_output=True.  This is a regression test for gh-4282.
+    y0 = 1
+    t = [0]
+    y, info = odeint(lambda y, t: -y, y0, t, full_output=True)
+    assert_array_equal(y, np.array([[y0]]))
+
+
 def test_odeint_banded_jacobian():
     # Test the use of the `Dfun`, `ml` and `mu` options of odeint.
 
@@ -552,42 +625,117 @@ def test_odeint_banded_jacobian():
     def jac(y, t, c):
         return c
 
-    def bjac_cols(y, t, c):
-        return np.column_stack((np.r_[0, np.diag(c, 1)], np.diag(c)))
+    def jac_transpose(y, t, c):
+        return c.T.copy(order='C')
 
     def bjac_rows(y, t, c):
-        return np.row_stack((np.r_[0, np.diag(c, 1)], np.diag(c)))
+        jac = np.row_stack((np.r_[0, np.diag(c, 1)],
+                            np.diag(c),
+                            np.r_[np.diag(c, -1), 0],
+                            np.r_[np.diag(c, -2), 0, 0]))
+        return jac
 
-    c = array([[-50,   75,     0],
-               [  0, -0.1,     1],
-               [  0,    0, -1e-4]])
+    def bjac_cols(y, t, c):
+        return bjac_rows(y, t, c).T.copy(order='C')
 
-    y0 = arange(3)
-    t = np.linspace(0, 50, 6)
+    c = array([[-205, 0.01, 0.00, 0.0],
+               [0.1, -2.50, 0.02, 0.0],
+               [1e-3, 0.01, -2.0, 0.01],
+               [0.00, 0.00, 0.1, -1.0]])
 
-    # The results of the following three calls should be the same.
-    sol0, info0 = odeint(func, y0, t, args=(c,), full_output=True,
+    y0 = np.ones(4)
+    t = np.array([0, 5, 10, 100])
+
+    # Use the full Jacobian.
+    sol1, info1 = odeint(func, y0, t, args=(c,), full_output=True,
+                         atol=1e-13, rtol=1e-11, mxstep=10000,
                          Dfun=jac)
 
-    sol1, info1 = odeint(func, y0, t, args=(c,), full_output=True,
-                         Dfun=bjac_cols, ml=0, mu=1, col_deriv=True)
-
+    # Use the transposed full Jacobian, with col_deriv=True.
     sol2, info2 = odeint(func, y0, t, args=(c,), full_output=True,
-                         Dfun=bjac_rows, ml=0, mu=1)
+                         atol=1e-13, rtol=1e-11, mxstep=10000,
+                         Dfun=jac_transpose, col_deriv=True)
 
-    # These could probably be compared using `assert_array_equal`.
-    # The code paths might not be *exactly* the same, so `allclose` is used
-    # to compare the solutions.
-    assert_allclose(sol0, sol1)
-    assert_allclose(sol0, sol2)
+    # Use the banded Jacobian.
+    sol3, info3 = odeint(func, y0, t, args=(c,), full_output=True,
+                         atol=1e-13, rtol=1e-11, mxstep=10000,
+                         Dfun=bjac_rows, ml=2, mu=1)
 
-    # Verify that the number of jacobian evaluations was the same
-    # for all three calls of odeint.  This is a regression test--there
-    # was a bug in the handling of banded jacobians that resulted in
-    # an incorrect jacobian matrix being passed to the LSODA code.
-    # That would cause errors or excessive jacobian evaluations.
-    assert_array_equal(info0['nje'], info1['nje'])
-    assert_array_equal(info0['nje'], info2['nje'])
+    # Use the transposed banded Jacobian, with col_deriv=True.
+    sol4, info4 = odeint(func, y0, t, args=(c,), full_output=True,
+                         atol=1e-13, rtol=1e-11, mxstep=10000,
+                         Dfun=bjac_cols, ml=2, mu=1, col_deriv=True)
+
+    assert_allclose(sol1, sol2, err_msg="sol1 != sol2")
+    assert_allclose(sol1, sol3, atol=1e-12, err_msg="sol1 != sol3")
+    assert_allclose(sol3, sol4, err_msg="sol3 != sol4")
+
+    # Verify that the number of jacobian evaluations was the same for the
+    # calls of odeint with a full jacobian and with a banded jacobian. This is
+    # a regression test--there was a bug in the handling of banded jacobians
+    # that resulted in an incorrect jacobian matrix being passed to the LSODA
+    # code.  That would cause errors or excessive jacobian evaluations.
+    assert_array_equal(info1['nje'], info2['nje'])
+    assert_array_equal(info3['nje'], info4['nje'])
+
+
+def test_odeint_errors():
+    def sys1d(x, t):
+        return -100*x
+
+    def bad1(x, t):
+        return 1.0/0
+
+    def bad2(x, t):
+        return "foo"
+
+    def bad_jac1(x, t):
+        return 1.0/0
+
+    def bad_jac2(x, t):
+        return [["foo"]]
+
+    def sys2d(x, t):
+        return [-100*x[0], -0.1*x[1]]
+
+    def sys2d_bad_jac(x, t):
+        return [[1.0/0, 0], [0, -0.1]]
+
+    assert_raises(ZeroDivisionError, odeint, bad1, 1.0, [0, 1])
+    assert_raises(ValueError, odeint, bad2, 1.0, [0, 1])
+
+    assert_raises(ZeroDivisionError, odeint, sys1d, 1.0, [0, 1], Dfun=bad_jac1)
+    assert_raises(ValueError, odeint, sys1d, 1.0, [0, 1], Dfun=bad_jac2)
+
+    assert_raises(ZeroDivisionError, odeint, sys2d, [1.0, 1.0], [0, 1],
+                  Dfun=sys2d_bad_jac)
+
+
+def test_odeint_bad_shapes():
+    # Tests of some errors that can occur with odeint.
+
+    def badrhs(x, t):
+        return [1, -1]
+
+    def sys1(x, t):
+        return -100*x
+
+    def badjac(x, t):
+        return [[0, 0, 0]]
+
+    # y0 must be at most 1-d.
+    bad_y0 = [[0, 0], [0, 0]]
+    assert_raises(ValueError, odeint, sys1, bad_y0, [0, 1])
+
+    # t must be at most 1-d.
+    bad_t = [[0, 1], [2, 3]]
+    assert_raises(ValueError, odeint, sys1, [10.0], bad_t)
+
+    # y0 is 10, but badrhs(x, t) returns [1, -1].
+    assert_raises(RuntimeError, odeint, badrhs, 10, [0, 1])
+
+    # shape of array returned by badjac(x, t) is not correct.
+    assert_raises(RuntimeError, odeint, sys1, [10, 10], [0, 1], Dfun=badjac)
 
 
 if __name__ == "__main__":
