@@ -13,35 +13,33 @@ from glob import glob
 from io import BytesIO
 from tempfile import mkdtemp
 
-from scipy.lib.six import u, text_type, string_types
+from scipy._lib.six import u, text_type, string_types
 
 import warnings
 import shutil
 import gzip
 
 from numpy.testing import (assert_array_equal, assert_array_almost_equal,
-                           assert_equal, assert_raises, run_module_suite)
-from numpy.testing.utils import WarningManager
-
-from nose.tools import assert_true
+                           assert_equal, assert_raises, run_module_suite,
+                           assert_)
 
 import numpy as np
 from numpy import array
 import scipy.sparse as SP
 
 import scipy.io.matlab.byteordercodes as boc
-from scipy.io.matlab.miobase import matdims, MatWriteError
+from scipy.io.matlab.miobase import matdims, MatWriteError, MatReadError
 from scipy.io.matlab.mio import (mat_reader_factory, loadmat, savemat, whosmat)
 from scipy.io.matlab.mio5 import (MatlabObject, MatFile5Writer, MatFile5Reader,
-                                  MatlabFunction, varmats_from_mat)
-
+                                  MatlabFunction, varmats_from_mat,
+                                  to_writeable, EmptyStructMarker)
+from scipy.io.matlab import mio5_params as mio5p
 
 test_data_path = pjoin(dirname(__file__), 'data')
 
 
 def mlarr(*args, **kwargs):
-    ''' Convenience function to return matlab-compatible 2D array
-    '''
+    """Convenience function to return matlab-compatible 2D array."""
     arr = np.array(*args, **kwargs)
     arr.shape = matdims(arr)
     return arr
@@ -233,10 +231,10 @@ case_table5_rt.append(
 
 
 def types_compatible(var1, var2):
-    ''' Check if types are same or compatible
+    """Check if types are same or compatible.
 
-    0d numpy scalars are compatible with bare python scalars
-    '''
+    0-D numpy scalars are compatible with bare python scalars.
+    """
     type1 = type(var1)
     type2 = type(var2)
     if type1 is type2:
@@ -251,16 +249,16 @@ def types_compatible(var1, var2):
 def _check_level(label, expected, actual):
     """ Check one level of a potentially nested array """
     if SP.issparse(expected):  # allow different types of sparse matrices
-        assert_true(SP.issparse(actual))
+        assert_(SP.issparse(actual))
         assert_array_almost_equal(actual.todense(),
                                   expected.todense(),
                                   err_msg=label,
                                   decimal=5)
         return
     # Check types are as expected
-    assert_true(types_compatible(expected, actual),
-           "Expected type %s, got %s at %s" %
-                (type(expected), type(actual), label))
+    assert_(types_compatible(expected, actual),
+            "Expected type %s, got %s at %s" %
+            (type(expected), type(actual), label))
     # A field in a record array may not be an ndarray
     # A scalar from a record array will be type np.void
     if not isinstance(expected,
@@ -268,11 +266,10 @@ def _check_level(label, expected, actual):
         assert_equal(expected, actual)
         return
     # This is an ndarray-like thing
-    assert_true(expected.shape == actual.shape,
-                msg='Expected shape %s, got %s at %s' % (expected.shape,
-                                                         actual.shape,
-                                                         label)
-                )
+    assert_(expected.shape == actual.shape,
+            msg='Expected shape %s, got %s at %s' % (expected.shape,
+                                                     actual.shape,
+                                                     label))
     ex_dtype = expected.dtype
     if ex_dtype.hasobject:  # array of objects
         if isinstance(expected, MatlabObject):
@@ -302,7 +299,7 @@ def _load_check_case(name, files, case):
         label = "test %s; file %s" % (name, file_name)
         for k, expected in case.items():
             k_label = "%s, variable %s" % (label, k)
-            assert_true(k in matdict, "Missing key at %s" % k_label)
+            assert_(k in matdict, "Missing key at %s" % k_label)
             _check_level(k_label, expected, matdict[k])
 
 
@@ -338,8 +335,8 @@ def test_load():
         expected = case['expected']
         filt = pjoin(test_data_path, 'test%s_*.mat' % name)
         files = glob(filt)
-        assert_true(len(files) > 0,
-                    "No files for test %s using filter %s" % (name, filt))
+        assert_(len(files) > 0,
+                "No files for test %s using filter %s" % (name, filt))
         yield _load_check_case, name, files, expected
 
 
@@ -351,8 +348,8 @@ def test_whos():
         classes = case['classes']
         filt = pjoin(test_data_path, 'test%s_*.mat' % name)
         files = glob(filt)
-        assert_true(len(files) > 0,
-                    "No files for test %s using filter %s" % (name, filt))
+        assert_(len(files) > 0,
+                "No files for test %s using filter %s" % (name, filt))
         yield _whos_check_case, name, files, expected, classes
 
 
@@ -427,7 +424,7 @@ def test_mat73():
     # Check any hdf5 files raise an error
     filenames = glob(
         pjoin(test_data_path, 'testhdf5*.mat'))
-    assert_true(len(filenames) > 0)
+    assert_(len(filenames) > 0)
     for filename in filenames:
         fp = open(filename, 'rb')
         assert_raises(NotImplementedError,
@@ -451,8 +448,15 @@ def test_warnings():
 
 
 def test_regression_653():
-    """Regression test for #653."""
-    assert_raises(TypeError, savemat, BytesIO(), {'d':{1:2}}, format='5')
+    # Saving a dictionary with only invalid keys used to raise an error. Now we
+    # save this as an empty struct in matlab space.
+    sio = BytesIO()
+    savemat(sio, {'d':{1:2}}, format='5')
+    back = loadmat(sio)['d']
+    # Check we got an empty struct equivalent
+    assert_equal(back.shape, (1,1))
+    assert_equal(back.dtype, np.dtype(np.object))
+    assert_(back[0,0] is None)
 
 
 def test_structname_len():
@@ -460,7 +464,6 @@ def test_structname_len():
     lim = 31
     fldname = 'a' * lim
     st1 = np.zeros((1,1), dtype=[(fldname, object)])
-    mat_stream = BytesIO()
     savemat(BytesIO(), {'longstruct': st1}, format='5')
     fldname = 'a' * (lim+1)
     st1 = np.zeros((1,1), dtype=[(fldname, object)])
@@ -480,7 +483,6 @@ def test_long_field_names():
     lim = 63
     fldname = 'a' * lim
     st1 = np.zeros((1,1), dtype=[(fldname, object)])
-    mat_stream = BytesIO()
     savemat(BytesIO(), {'longstruct': st1}, format='5',long_field_names=True)
     fldname = 'a' * (lim+1)
     st1 = np.zeros((1,1), dtype=[(fldname, object)])
@@ -497,7 +499,6 @@ def test_long_field_names_in_struct():
     st1 = np.zeros((1,1), dtype=[(fldname, object)])
     cell[0,0] = st1
     cell[0,1] = st1
-    mat_stream = BytesIO()
     savemat(BytesIO(), {'longstruct': cell}, format='5',long_field_names=True)
     #
     # Check to make sure it fails with long field names off
@@ -513,12 +514,10 @@ def test_cell_with_one_thing_in_it():
     cells = np.ndarray((1,2),dtype=object)
     cells[0,0] = 'Hello'
     cells[0,1] = 'World'
-    mat_stream = BytesIO()
     savemat(BytesIO(), {'x': cells}, format='5')
 
     cells = np.ndarray((1,1),dtype=object)
     cells[0,0] = 'Hello, world'
-    mat_stream = BytesIO()
     savemat(BytesIO(), {'x': cells}, format='5')
 
 
@@ -548,12 +547,12 @@ def test_use_small_element():
     sio.truncate(0)
     sio.seek(0)
     wtr.put_variables({'aaaa': arr})
-    yield assert_true, w_sz - len(sio.getvalue()) > 4
+    yield assert_, w_sz - len(sio.getvalue()) > 4
     # Whereas increasing name size makes less difference
     sio.truncate(0)
     sio.seek(0)
     wtr.put_variables({'aaaaaa': arr})
-    yield assert_true, len(sio.getvalue()) - w_sz < 4
+    yield assert_, len(sio.getvalue()) - w_sz < 4
 
 
 def test_save_dict():
@@ -622,7 +621,7 @@ def test_compression():
     compressed_len = len(stream.getvalue())
     vals = loadmat(stream)
     yield assert_array_equal, vals['arr'], arr
-    yield assert_true, raw_len > compressed_len
+    yield assert_, raw_len > compressed_len
     # Concatenate, test later
     arr2 = arr.copy()
     arr2[0,0] = 1
@@ -657,8 +656,8 @@ def test_skip_variable():
     # Prove that it loads with loadmat
     #
     d = loadmat(filename, struct_as_record=True)
-    yield assert_true, 'first' in d
-    yield assert_true, 'second' in d
+    yield assert_, 'first' in d
+    yield assert_, 'second' in d
     #
     # Make the factory
     #
@@ -667,7 +666,7 @@ def test_skip_variable():
     # This is where the factory breaks with an error in MatMatrixGetter.to_next
     #
     d = factory.get_variables('second')
-    yield assert_true, 'second' in d
+    yield assert_, 'second' in d
     factory.mat_stream.close()
 
 
@@ -680,7 +679,7 @@ def test_empty_struct():
     a = d['a']
     assert_equal(a.shape, (1,1))
     assert_equal(a.dtype, np.dtype(np.object))
-    assert_true(a[0,0] is None)
+    assert_(a[0,0] is None)
     stream = BytesIO()
     arr = np.array((), dtype='U')
     # before ticket fix, this used to give data type not understood
@@ -690,6 +689,91 @@ def test_empty_struct():
     assert_array_equal(a2, arr)
 
 
+def test_save_empty_dict():
+    # saving empty dict also gives empty struct
+    stream = BytesIO()
+    savemat(stream, {'arr': {}})
+    d = loadmat(stream)
+    a = d['arr']
+    assert_equal(a.shape, (1,1))
+    assert_equal(a.dtype, np.dtype(np.object))
+    assert_(a[0,0] is None)
+
+
+def assert_any_equal(output, alternatives):
+    """ Assert `output` is equal to at least one element in `alternatives`
+    """
+    one_equal = False
+    for expected in alternatives:
+        if np.all(output == expected):
+            one_equal = True
+            break
+    assert_(one_equal)
+
+
+def test_to_writeable():
+    # Test to_writeable function
+    res = to_writeable(np.array([1]))  # pass through ndarrays
+    assert_equal(res.shape, (1,))
+    assert_array_equal(res, 1)
+    # Dict fields can be written in any order
+    expected1 = np.array([(1, 2)], dtype=[('a', '|O8'), ('b', '|O8')])
+    expected2 = np.array([(2, 1)], dtype=[('b', '|O8'), ('a', '|O8')])
+    alternatives = (expected1, expected2)
+    assert_any_equal(to_writeable({'a':1,'b':2}), alternatives)
+    # Fields with underscores discarded
+    assert_any_equal(to_writeable({'a':1,'b':2, '_c':3}), alternatives)
+    # Not-string fields discarded
+    assert_any_equal(to_writeable({'a':1,'b':2, 100:3}), alternatives)
+    # String fields that are valid Python identifiers discarded
+    assert_any_equal(to_writeable({'a':1,'b':2, '99':3}), alternatives)
+    # Object with field names is equivalent
+
+    class klass(object):
+        pass
+
+    c = klass
+    c.a = 1
+    c.b = 2
+    assert_any_equal(to_writeable(c), alternatives)
+    # empty list and tuple go to empty array
+    res = to_writeable([])
+    assert_equal(res.shape, (0,))
+    assert_equal(res.dtype.type, np.float64)
+    res = to_writeable(())
+    assert_equal(res.shape, (0,))
+    assert_equal(res.dtype.type, np.float64)
+    # None -> None
+    assert_(to_writeable(None) is None)
+    # String to strings
+    assert_equal(to_writeable('a string').dtype.type, np.str_)
+    # Scalars to numpy to numpy scalars
+    res = to_writeable(1)
+    assert_equal(res.shape, ())
+    assert_equal(res.dtype.type, np.array(1).dtype.type)
+    assert_array_equal(res, 1)
+    # Empty dict returns EmptyStructMarker
+    assert_(to_writeable({}) is EmptyStructMarker)
+    # Object does not have (even empty) __dict__
+    assert_(to_writeable(object()) is None)
+    # Custom object does have empty __dict__, returns EmptyStructMarker
+
+    class C(object):
+        pass
+
+    assert_(to_writeable(c()) is EmptyStructMarker)
+    # dict keys with legal characters are convertible
+    res = to_writeable({'a': 1})['a']
+    assert_equal(res.shape, (1,))
+    assert_equal(res.dtype.type, np.object_)
+    # Only fields with illegal characters, falls back to EmptyStruct
+    assert_(to_writeable({'1':1}) is EmptyStructMarker)
+    assert_(to_writeable({'_a':1}) is EmptyStructMarker)
+    # Unless there are valid fields, in which case structured array
+    assert_equal(to_writeable({'1':1, 'f': 2}),
+                 np.array([(2,)], dtype=[('f', '|O8')]))
+
+
 def test_recarray():
     # check roundtrip of structured array
     dt = [('f1', 'f8'),
@@ -797,6 +881,21 @@ def test_empty_string():
     stream.close()
 
 
+def test_corrupted_data():
+    import zlib
+    for exc, fname in [(ValueError, 'corrupted_zlib_data.mat'),
+                       (zlib.error, 'corrupted_zlib_checksum.mat')]:
+        with open(pjoin(test_data_path, fname), 'rb') as fp:
+            rdr = MatFile5Reader(fp)
+            assert_raises(exc, rdr.get_variables)
+
+
+def test_corrupted_data_check_can_be_disabled():
+    with open(pjoin(test_data_path, 'corrupted_zlib_data.mat'), 'rb') as fp:
+        rdr = MatFile5Reader(fp, verify_compressed_data_integrity=False)
+        rdr.get_variables()
+
+
 def test_read_both_endian():
     # make sure big- and little- endian data is read correctly
     for fname in ('big_endian.mat', 'little_endian.mat'):
@@ -832,17 +931,33 @@ def test_write_opposite_endian():
 
 
 def test_logical_array():
-    # The roundtrip test doesn't verify that we load the data up with the correct (bool) dtype
-    fp = open(pjoin(test_data_path, 'testbool_8_WIN64.mat'), 'rb')
-    rdr = MatFile5Reader(fp, mat_dtype=True)
-    d = rdr.get_variables()
-    fp.close()
-
+    # The roundtrip test doesn't verify that we load the data up with the
+    # correct (bool) dtype
+    with open(pjoin(test_data_path, 'testbool_8_WIN64.mat'), 'rb') as fobj:
+        rdr = MatFile5Reader(fobj, mat_dtype=True)
+        d = rdr.get_variables()
     x = np.array([[True], [False]], dtype=np.bool_)
     assert_array_equal(d['testbools'], x)
     assert_equal(d['testbools'].dtype, x.dtype)
 
 
+def test_logical_out_type():
+    # Confirm that bool type written as uint8, uint8 class
+    # See gh-4022
+    stream = BytesIO()
+    barr = np.array([False, True, False])
+    savemat(stream, {'barray': barr})
+    stream.seek(0)
+    reader = MatFile5Reader(stream)
+    reader.initialize_read()
+    reader.read_file_header()
+    hdr, _ = reader.read_var_header()
+    assert_equal(hdr.mclass, mio5p.mxUINT8_CLASS)
+    assert_equal(hdr.is_logical, True)
+    var = reader.read_var_array(hdr, False)
+    assert_equal(var.dtype.type, np.uint8)
+
+
 def test_mat4_3d():
     # test behavior when writing 3D arrays to matlab 4 files
     stream = BytesIO()
@@ -856,7 +971,7 @@ def test_func_read():
     rdr = MatFile5Reader(fp)
     d = rdr.get_variables()
     fp.close()
-    assert_true(isinstance(d['testfunc'], MatlabFunction))
+    assert_(isinstance(d['testfunc'], MatlabFunction))
     stream = BytesIO()
     wtr = MatFile5Writer(stream)
     assert_raises(MatWriteError, wtr.put_variables, d)
@@ -905,9 +1020,9 @@ def test_scalar_squeeze():
     in_d = {'scalar': [[0.1]], 'string': 'my name', 'st':{'one':1, 'two':2}}
     savemat(stream, in_d)
     out_d = loadmat(stream, squeeze_me=True)
-    assert_true(isinstance(out_d['scalar'], float))
-    assert_true(isinstance(out_d['string'], string_types))
-    assert_true(isinstance(out_d['st'], np.ndarray))
+    assert_(isinstance(out_d['scalar'], float))
+    assert_(isinstance(out_d['string'], string_types))
+    assert_(isinstance(out_d['st'], np.ndarray))
 
 
 def test_str_round():
@@ -954,8 +1069,8 @@ def test_loadmat_varnames():
                       '__header__',
                       '__version__']
     for eg_file, sys_v_names in (
-        (pjoin(test_data_path, 'testmulti_4.2c_SOL2.mat'), []),
-        (pjoin(test_data_path, 'testmulti_7.4_GLNX86.mat'), mat5_sys_names)):
+        (pjoin(test_data_path, 'testmulti_4.2c_SOL2.mat'), []), (pjoin(
+            test_data_path, 'testmulti_7.4_GLNX86.mat'), mat5_sys_names)):
         vars = loadmat(eg_file)
         assert_equal(set(vars.keys()), set(['a', 'theta'] + sys_v_names))
         vars = loadmat(eg_file, variable_names='a')
@@ -1007,7 +1122,7 @@ def test_varmats_from_mat():
 
 
 def test_one_by_zero():
-    ''' Test 1x0 chars get read correctly '''
+    # Test 1x0 chars get read correctly
     func_eg = pjoin(test_data_path, 'one_by_zero_char.mat')
     fp = open(func_eg, 'rb')
     rdr = MatFile5Reader(fp)
@@ -1033,16 +1148,82 @@ def test_unicode_mat4():
     assert_equal(var_back['second_cat'], var['second_cat'])
 
 
+def test_logical_sparse():
+    # Test we can read logical sparse stored in mat file as bytes.
+    # See https://github.com/scipy/scipy/issues/3539.
+    # In some files saved by MATLAB, the sparse data elements (Real Part
+    # Subelement in MATLAB speak) are stored with apparent type double
+    # (miDOUBLE) but are in fact single bytes.
+    filename = pjoin(test_data_path,'logical_sparse.mat')
+    # Before fix, this would crash with:
+    # ValueError: indices and data should have the same size
+    d = loadmat(filename, struct_as_record=True)
+    log_sp = d['sp_log_5_4']
+    assert_(isinstance(log_sp, SP.csc_matrix))
+    assert_equal(log_sp.dtype.type, np.bool_)
+    assert_array_equal(log_sp.toarray(),
+                       [[True, True, True, False],
+                        [False, False, True, False],
+                        [False, False, True, False],
+                        [False, False, False, False],
+                        [False, False, False, False]])
+
+
 def test_empty_sparse():
     # Can we read empty sparse matrices?
     sio = BytesIO()
     import scipy.sparse
     empty_sparse = scipy.sparse.csr_matrix([[0,0],[0,0]])
-    savemat(sio, dict(x = empty_sparse))
+    savemat(sio, dict(x=empty_sparse))
     sio.seek(0)
     res = loadmat(sio)
     assert_array_equal(res['x'].shape, empty_sparse.shape)
     assert_array_equal(res['x'].todense(), 0)
+    # Do empty sparse matrices get written with max nnz 1?
+    # See https://github.com/scipy/scipy/issues/4208
+    sio.seek(0)
+    reader = MatFile5Reader(sio)
+    reader.initialize_read()
+    reader.read_file_header()
+    hdr, _ = reader.read_var_header()
+    assert_equal(hdr.nzmax, 1)
+
+
+def test_empty_mat_error():
+    # Test we get a specific warning for an empty mat file
+    sio = BytesIO()
+    assert_raises(MatReadError, loadmat, sio)
+
+
+def test_miuint32_compromise():
+    # Reader should accept miUINT32 for miINT32, but check signs
+    # mat file with miUINT32 for miINT32, but OK values
+    filename = pjoin(test_data_path, 'miuint32_for_miint32.mat')
+    res = loadmat(filename)
+    assert_equal(res['an_array'], np.arange(10)[None, :])
+    # mat file with miUINT32 for miINT32, with negative value
+    filename = pjoin(test_data_path, 'bad_miuint32.mat')
+    with warnings.catch_warnings(record=True):  # Py3k ResourceWarning
+        assert_raises(ValueError, loadmat, filename)
+
+
+def test_miutf8_for_miint8_compromise():
+    # Check reader accepts ascii as miUTF8 for array names
+    filename = pjoin(test_data_path, 'miutf8_array_name.mat')
+    res = loadmat(filename)
+    assert_equal(res['array_name'], [[1]])
+    # mat file with non-ascii utf8 name raises error
+    filename = pjoin(test_data_path, 'bad_miutf8_array_name.mat')
+    with warnings.catch_warnings(record=True):  # Py3k ResourceWarning
+        assert_raises(ValueError, loadmat, filename)
+
+
+def test_bad_utf8():
+    # Check that reader reads bad UTF with 'replace' option
+    filename = pjoin(test_data_path,'broken_utf8.mat')
+    res = loadmat(filename)
+    assert_equal(res['bad_string'],
+                 b'\x80 am broken'.decode('utf8', 'replace'))
 
 
 if __name__ == "__main__":
