@@ -41,216 +41,6 @@ typedef struct {
     void* next;
 } _index_pair;
 
-#define CASE_LABEL(_p, _pi, _type) \
-case t ## _type:                   \
-    *_p = *(_type*)_pi ? -1 : 0;     \
-    break
-
-int NI_Label(PyArrayObject* input, PyArrayObject* strct,
-                         npy_intp *max_label, PyArrayObject* output)
-{
-    int kk;
-    npy_intp jj, ll, ssize, size, filter_size, *offsets = NULL;
-    npy_intp mask_value, *oo;
-    Bool *ps, *footprint = NULL;
-    char *pi, *po;
-    Int32 index = 0, *index_map = NULL;
-    NI_FilterIterator fi;
-    NI_Iterator ii, io;
-    _index_pair *pairs = NULL;
-
-    /* structure size */
-    ssize = 1;
-    for(kk = 0; kk < strct->nd; kk++)
-        ssize *= strct->dimensions[kk];
-    /* we only use the first half of the structure data, so we make a
-         temporary structure for use with the filter functions: */
-    footprint = (Bool*)malloc(ssize * sizeof(Bool));
-    if (!footprint) {
-        PyErr_NoMemory();
-        goto exit;
-    }
-    ps = (Bool*)PyArray_DATA(strct);
-    filter_size = 0;
-    for(jj = 0; jj < ssize / 2; jj++) {
-        footprint[jj] = ps[jj];
-        if (ps[jj])
-            ++filter_size;
-    }
-    for(jj = ssize / 2; jj < ssize; jj++)
-        footprint[jj] = 0;
-    /* get data and size */
-    pi = (void *)PyArray_DATA(input);
-    po = (void *)PyArray_DATA(output);
-    size = 1;
-    for(kk = 0; kk < output->nd; kk++)
-        size *= output->dimensions[kk];
-    if (!NI_InitPointIterator(input, &ii))
-        goto exit;
-    if (!NI_InitPointIterator(output, &io))
-        goto exit;
-    /* set all elements in the output corresponding to non-zero elements
-         in input to -1: */
-    for(jj = 0; jj < size; jj++) {
-        Int32 *p = (Int32*)po;
-        switch (input->descr->type_num) {
-        CASE_LABEL(p, pi, Bool);
-        CASE_LABEL(p, pi, UInt8);
-        CASE_LABEL(p, pi, UInt16);
-        CASE_LABEL(p, pi, UInt32);
-#if HAS_UINT64
-        CASE_LABEL(p, pi, UInt64);
-#endif
-        CASE_LABEL(p, pi, Int8);
-        CASE_LABEL(p, pi, Int16);
-        CASE_LABEL(p, pi, Int32);
-        CASE_LABEL(p, pi, Int64);
-        CASE_LABEL(p, pi, Float32);
-        CASE_LABEL(p, pi, Float64);
-        default:
-            PyErr_SetString(PyExc_RuntimeError, "data type not supported");
-            goto exit;
-        }
-        NI_ITERATOR_NEXT2(ii, io, pi, po);
-    }
-
-    /* calculate the filter offsets: */
-    if (!NI_InitFilterOffsets(output, footprint, strct->dimensions, NULL,
-                                                    NI_EXTEND_CONSTANT, &offsets, &mask_value, NULL))
-        goto exit;
-    /* initialize filter iterator: */
-    if (!NI_InitFilterIterator(input->nd, strct->dimensions, filter_size,
-                                                                                     input->dimensions, NULL, &fi))
-        goto exit;
-    /* reset output iterator: */
-    NI_ITERATOR_RESET(io);
-    po = (void *)PyArray_DATA(output);
-    /* iterator over the elements: */
-    oo = offsets;
-    for(jj = 0; jj < size; jj++) {
-        if (*(Int32*)po < 0) {
-            Int32 neighbor = 0;
-            /* iterate over structuring element: */
-            for(ll = 0; ll < filter_size; ll++) {
-                int offset = oo[ll];
-                if (offset != mask_value) {
-                    Int32 tt = *(Int32*)(po + offset);
-                    if (tt > 0) {
-                        /* this element is next to an already found object: */
-                        if (neighbor && neighbor != tt) {
-                            /* we have two objects that must be merged later: */
-                            _index_pair* tp = (_index_pair*)malloc(sizeof(_index_pair));
-                            if (!tp) {
-                                PyErr_NoMemory();
-                                goto exit;
-                            }
-                            tp->next = pairs;
-                            /* the pairs must be ordered: */
-                            if (neighbor < tt) {
-                                tp->index1 = neighbor;
-                                tp->index2 = tt;
-                            } else {
-                                tp->index1 = tt;
-                                tp->index2 = neighbor;
-                            }
-                            pairs = tp;
-                        } else {
-                            neighbor = tt;
-                        }
-                    }
-                }
-            }
-            if (neighbor) {
-                /* this point belongs to an existing object */
-                *(Int32*)po = neighbor;
-            } else {
-                /* this may be a new object: */
-                *(Int32*)po = ++index;
-            }
-        }
-        NI_FILTER_NEXT(fi, io, oo, po);
-    }
-    *max_label = index;
-    /* merge any touching objects: */
-    if (pairs) {
-        Int32 counter;
-        index_map = (Int32*)malloc(index * sizeof(Int32));
-        if (!index_map) {
-            PyErr_NoMemory();
-            goto exit;
-        }
-        for(jj = 0; jj < index; jj++)
-            index_map[jj] = (Int32)jj;
-        while (pairs) {
-            Int32 idx1 = pairs->index1 - 1;
-            Int32 idx2 = pairs->index2 - 1;
-            if (index_map[idx2] == idx1 || index_map[idx2] == idx2) {
-                /* if this pair was already processed, or if idx2 was not
-                     mapped yet, we delete this pair and map idx2 to idx1: */
-                _index_pair *tp = pairs;
-                pairs = tp->next;
-                free(tp);
-                index_map[idx2] = idx1;
-            } else {
-                /* idx2 was already mapped, therefore we find what it was
-                     mapped to and change the current pair to the result of
-                     that and idx1. Since the pair is not destroyed, it will be
-                     re-processed with the adapted values.  */
-                idx2 = index_map[idx2];
-                /* keep the pairs ordered: */
-                if (idx1 < idx2) {
-                    pairs->index1 = idx1 + 1;
-                    pairs->index2 = idx2 + 1;
-                } else {
-                    pairs->index1 = idx2 + 1;
-                    pairs->index2 = idx1 + 1;
-                }
-            }
-        }
-        for(jj = 0; jj < index; jj++) {
-            /* if the current index maps to a index that is also mapped,
-                 change it to map to that index. Since an index always maps to
-                 a lower index or to itself, this will make sure that at the
-                 end all indices map to an unmapped index. */
-            if (index_map[index_map[jj]] < index_map[jj])
-                index_map[jj] = index_map[index_map[jj]];
-        }
-        /* renumber the indices that are not mapped: */
-        counter = 0;
-        for(jj = 0; jj < index; jj++)
-            if (index_map[jj] == jj)
-                index_map[jj] = ++counter;
-            else
-                index_map[jj] = index_map[index_map[jj]];
-    }
-
-    /* relabel the output if we merged some objects: */
-    if (index_map) {
-        *max_label = 0;
-        NI_ITERATOR_RESET(io);
-        po = (void *)PyArray_DATA(output);
-        for(jj = 0; jj < size; jj++) {
-            Int32 p = *(Int32*)po;
-            if (p > 0 )
-                *(Int32*)po = index_map[p - 1];
-            if (*(Int32*)po > *max_label)
-                *max_label = *(Int32*)po;
-            NI_ITERATOR_NEXT(io, po);
-        }
-    }
- exit:
-    if (offsets) free(offsets);
-    if (index_map) free(index_map);
-    while (pairs) {
-        _index_pair *tp = pairs;
-        pairs = (_index_pair*)pairs->next;
-        free(tp);
-    }
-    if (footprint)
-        free(footprint);
-    return PyErr_Occurred() ? 0 : 1;
-}
-
 #define CASE_FIND_OBJECT_POINT(_pi, _regions, _rank, _dimensions, \
                                                              _max_label, _ii, _type)            \
 case t ## _type:                                                  \
@@ -289,6 +79,9 @@ int NI_FindObjects(PyArrayObject* input, npy_intp max_label,
     npy_intp size, jj;
     NI_Iterator ii;
     char *pi;
+    NPY_BEGIN_THREADS_DEF;
+
+    NPY_BEGIN_THREADS;
 
     /* get input data, size and iterator: */
     pi = (void *)PyArray_DATA(input);
@@ -306,7 +99,7 @@ int NI_FindObjects(PyArrayObject* input, npy_intp max_label,
     }
     /* iterate over all points: */
     for(jj = 0 ; jj < size; jj++) {
-        switch (input->descr->type_num) {
+        switch (NI_NormalizeType(input->descr->type_num)) {
         CASE_FIND_OBJECT_POINT(pi, regions, input->nd, input->dimensions,
                                                      max_label, ii,  Bool);
         CASE_FIND_OBJECT_POINT(pi, regions, input->nd, input->dimensions,
@@ -329,12 +122,14 @@ int NI_FindObjects(PyArrayObject* input, npy_intp max_label,
                                                      max_label, ii, Int64);
             break;
         default:
+            NPY_END_THREADS;
             PyErr_SetString(PyExc_RuntimeError, "data type not supported");
             goto exit;
         }
         NI_ITERATOR_NEXT(ii, pi);
     }
  exit:
+    NPY_END_THREADS;
     return PyErr_Occurred() ? 0 : 1;
 }
 
@@ -378,6 +173,7 @@ int NI_FindObjects(PyArrayObject* input, npy_intp max_label,
         _v = *(Float64*)_pi;                                              \
         break;                                                            \
     default:                                                            \
+            NPY_END_THREADS;                                          \
             PyErr_SetString(PyExc_RuntimeError,                       \
                                          "data type not supported");  \
             return 0;                                                       \
@@ -418,6 +214,7 @@ int NI_FindObjects(PyArrayObject* input, npy_intp max_label,
         _v = *(Float64*)_pi;                                              \
         break;                                                            \
     default:                                                            \
+            NPY_END_THREADS;                                          \
             PyErr_SetString(PyExc_RuntimeError,                       \
                                         "data type not supported");   \
             return 0;                                                       \
@@ -465,6 +262,7 @@ int NI_FindObjects(PyArrayObject* input, npy_intp max_label,
             _label = *(Float64*)_pm;                                        \
             break;                                                          \
         default:                                                          \
+            NPY_END_THREADS;                                          \
             PyErr_SetString(PyExc_RuntimeError,                       \
                                         "data type not supported");   \
             return 0;                                                       \
@@ -507,6 +305,7 @@ int NI_FindObjects(PyArrayObject* input, npy_intp max_label,
             _label = *(Float64*)_pm;                                        \
             break;                                                          \
         default:                                                          \
+            NPY_END_THREADS;                                          \
             PyErr_SetString(PyExc_RuntimeError,                       \
                                         "data type not supported");   \
             return 0;                                                       \
@@ -524,6 +323,7 @@ int NI_Statistics(PyArrayObject *input, PyArrayObject *labels,
     NI_Iterator ii, mi;
     npy_intp jj, size, idx = 0, label = 1, doit = 1;
     int qq;
+    NPY_BEGIN_THREADS_DEF;
 
     /* input iterator: */
     if (!NI_InitPointIterator(input, &ii))
@@ -535,6 +335,9 @@ int NI_Statistics(PyArrayObject *input, PyArrayObject *labels,
             return 0;
         pm = (void *)PyArray_DATA(labels);
     }
+
+    NPY_BEGIN_THREADS;
+
     /* input size: */
     size = 1;
     for(qq = 0; qq < input->nd; qq++)
@@ -557,7 +360,7 @@ int NI_Statistics(PyArrayObject *input, PyArrayObject *labels,
     }
     /* iterate over array: */
     for(jj = 0; jj < size; jj++) {
-        NI_GET_LABEL(pm, label, labels->descr->type_num);
+        NI_GET_LABEL(pm, label, NI_NormalizeType(labels->descr->type_num));
         if (min_label >= 0) {
             if (label >= min_label && label <= max_label) {
                 idx = indices[label - min_label];
@@ -570,7 +373,7 @@ int NI_Statistics(PyArrayObject *input, PyArrayObject *labels,
         }
         if (doit) {
             double val;
-            NI_GET_VALUE(pi, val, input->descr->type_num);
+            NI_GET_VALUE(pi, val, NI_NormalizeType(input->descr->type_num));
             if (sum)
                 sum[idx] += val;
             if (total)
@@ -621,7 +424,7 @@ int NI_Statistics(PyArrayObject *input, PyArrayObject *labels,
                 pm = (void *)PyArray_DATA(labels);
             }
             for(jj = 0; jj < size; jj++) {
-                NI_GET_LABEL(pm, label, labels->descr->type_num);
+                NI_GET_LABEL(pm, label, NI_NormalizeType(labels->descr->type_num));
                 if (min_label >= 0) {
                     if (label >= min_label && label <= max_label) {
                         idx = indices[label - min_label];
@@ -634,7 +437,7 @@ int NI_Statistics(PyArrayObject *input, PyArrayObject *labels,
                 }
                 if (doit) {
                     double val;
-                    NI_GET_VALUE(pi, val, input->descr->type_num);
+                    NI_GET_VALUE(pi, val, NI_NormalizeType(input->descr->type_num));
                     val = val - sum[idx] / total[idx];
                     variance[idx] += val * val;
                 }
@@ -649,6 +452,7 @@ int NI_Statistics(PyArrayObject *input, PyArrayObject *labels,
                                                 variance[jj] / (total[jj] - 1) : 0.0);
         }
     }
+    NPY_END_THREADS;
     return 1;
 }
 
@@ -662,6 +466,7 @@ int NI_CenterOfMass(PyArrayObject *input, PyArrayObject *labels,
     npy_intp jj, kk, size, idx = 0, label = 1, doit = 1;
     double *sum = NULL;
     int qq;
+    NPY_BEGIN_THREADS_DEF;
 
     /* input iterator: */
     if (!NI_InitPointIterator(input, &ii))
@@ -677,11 +482,14 @@ int NI_CenterOfMass(PyArrayObject *input, PyArrayObject *labels,
     size = 1;
     for(qq = 0; qq < input->nd; qq++)
         size *= input->dimensions[qq];
-    sum = (double*)malloc(n_results * sizeof(double));
+    sum = malloc(n_results * sizeof(double));
     if (!sum) {
         PyErr_NoMemory();
         goto exit;
     }
+
+    NPY_BEGIN_THREADS;
+
     for(jj = 0; jj < n_results; jj++) {
         sum[jj] = 0.0;
         for(kk = 0; kk < input->nd; kk++)
@@ -689,7 +497,7 @@ int NI_CenterOfMass(PyArrayObject *input, PyArrayObject *labels,
     }
     /* iterate over array: */
     for(jj = 0; jj < size; jj++) {
-        NI_GET_LABEL(pm, label, labels->descr->type_num);
+        NI_GET_LABEL(pm, label, NI_NormalizeType(labels->descr->type_num));
         if (min_label >= 0) {
             if (label >= min_label && label <= max_label) {
                 idx = indices[label - min_label];
@@ -702,7 +510,7 @@ int NI_CenterOfMass(PyArrayObject *input, PyArrayObject *labels,
         }
         if (doit) {
             double val;
-            NI_GET_VALUE(pi, val, input->descr->type_num);
+            NI_GET_VALUE(pi, val, NI_NormalizeType(input->descr->type_num));
             sum[idx] += val;
             for(kk = 0; kk < input->nd; kk++)
                 center_of_mass[idx * input->nd + kk] += val * ii.coordinates[kk];
@@ -717,8 +525,8 @@ int NI_CenterOfMass(PyArrayObject *input, PyArrayObject *labels,
         for(kk = 0; kk < input->nd; kk++)
             center_of_mass[jj * input->nd + kk] /= sum[jj];
  exit:
-    if (sum)
-        free(sum);
+    NPY_END_THREADS;
+    free(sum);
     return  PyErr_Occurred() == NULL;
 }
 
@@ -734,6 +542,7 @@ int NI_Histogram(PyArrayObject *input, PyArrayObject *labels,
     Int32 **ph = NULL;
     double bsize;
     int qq;
+    NPY_BEGIN_THREADS_DEF;
 
     /* input iterator: */
     if (!NI_InitPointIterator(input, &ii))
@@ -745,11 +554,14 @@ int NI_Histogram(PyArrayObject *input, PyArrayObject *labels,
             goto exit;
         pm = (void *)PyArray_DATA(labels);
     }
-    ph = (Int32**)malloc(n_results * sizeof(Int32*));
+    ph = malloc(n_results * sizeof(Int32*));
     if (!ph) {
         PyErr_NoMemory();
         goto exit;
     }
+
+    NPY_BEGIN_THREADS;
+
     for(jj = 0; jj < n_results; jj++) {
             ph[jj] = (Int32*)PyArray_DATA(histograms[jj]);
             for(kk = 0; kk < nbins; kk++)
@@ -762,7 +574,7 @@ int NI_Histogram(PyArrayObject *input, PyArrayObject *labels,
         size *= input->dimensions[qq];
     /* iterate over array: */
     for(jj = 0; jj < size; jj++) {
-        NI_GET_LABEL(pm, label, labels->descr->type_num);
+        NI_GET_LABEL(pm, label, NI_NormalizeType(labels->descr->type_num));
         if (min_label >= 0) {
             if (label >= min_label && label <= max_label) {
                 idx = indices[label - min_label];
@@ -774,11 +586,11 @@ int NI_Histogram(PyArrayObject *input, PyArrayObject *labels,
             doit = label != 0;
         }
         if (doit) {
-            int bin;
+            npy_intp bin;
             double val;
-            NI_GET_VALUE(pi, val, input->descr->type_num);
+            NI_GET_VALUE(pi, val, NI_NormalizeType(input->descr->type_num));
             if (val >= min && val < max) {
-                bin = (int)((val - min) / bsize);
+                bin = (npy_intp)((val - min) / bsize);
                 ++(ph[idx][bin]);
             }
         }
@@ -789,8 +601,8 @@ int NI_Histogram(PyArrayObject *input, PyArrayObject *labels,
         }
     }
  exit:
-    if (ph)
-        free(ph);
+    NPY_END_THREADS;
+    free(ph);
     return  PyErr_Occurred() == NULL;
 }
 
@@ -876,6 +688,7 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
     NI_WatershedElement *temp = NULL, **first = NULL, **last = NULL;
     Bool *ps = NULL;
     NI_Iterator mi, ii, li;
+    NPY_BEGIN_THREADS_DEF;
 
     i_contiguous = PyArray_ISCONTIGUOUS(input);
     o_contiguous = PyArray_ISCONTIGUOUS(output);
@@ -890,22 +703,26 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
     for(ll = 0; ll < input->nd; ll++)
         size *= input->dimensions[ll];
     /* Storage for the temporary queue data. */
-    temp = (NI_WatershedElement*)malloc(size * sizeof(NI_WatershedElement));
+    temp = malloc(size * sizeof(NI_WatershedElement));
     if (!temp) {
         PyErr_NoMemory();
         goto exit;
     }
+
+    NPY_BEGIN_THREADS;
+
     pi = (void *)PyArray_DATA(input);
     if (!NI_InitPointIterator(input, &ii))
         goto exit;
     /* Initialization and find the maximum of the input. */
     maxval = 0;
     for(jj = 0; jj < size; jj++) {
-        int ival = 0;
-        switch(input->descr->type_num) {
+        npy_intp ival = 0;
+        switch(NI_NormalizeType(input->descr->type_num)) {
         CASE_GET_INPUT(ival, pi, UInt8);
         CASE_GET_INPUT(ival, pi, UInt16);
         default:
+            NPY_END_THREADS;
             PyErr_SetString(PyExc_RuntimeError, "data type not supported");
             goto exit;
         }
@@ -917,11 +734,10 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
     }
     pi = (void *)PyArray_DATA(input);
     /* Allocate and initialize the storage for the queue. */
-    first = (NI_WatershedElement**)malloc((maxval + 1) *
-                                                                                sizeof(NI_WatershedElement*));
-    last = (NI_WatershedElement**)malloc((maxval + 1) *
-                                                                             sizeof(NI_WatershedElement*));
-    if (!first || !last) {
+    first = malloc((maxval + 1) * sizeof(NI_WatershedElement*));
+    last = malloc((maxval + 1) * sizeof(NI_WatershedElement*));
+    if (NI_UNLIKELY(!first || !last)) {
+        NPY_END_THREADS;
         PyErr_NoMemory();
         goto exit;
     }
@@ -941,7 +757,7 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
     for(jj = 0; jj < size; jj++) {
         /* get marker */
         int label = 0;
-        switch(markers->descr->type_num) {
+        switch(NI_NormalizeType(markers->descr->type_num)) {
         CASE_GET_LABEL(label, pm, UInt8);
         CASE_GET_LABEL(label, pm, UInt16);
         CASE_GET_LABEL(label, pm, UInt32);
@@ -953,10 +769,11 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
         CASE_GET_LABEL(label, pm, Int32);
         CASE_GET_LABEL(label, pm, Int64);
         default:
+            NPY_END_THREADS;
             PyErr_SetString(PyExc_RuntimeError, "data type not supported");
             goto exit;
         }
-        switch(output->descr->type_num) {
+        switch(NI_NormalizeType(output->descr->type_num)) {
         CASE_PUT_LABEL(label, pl, UInt8);
         CASE_PUT_LABEL(label, pl, UInt16);
         CASE_PUT_LABEL(label, pl, UInt32);
@@ -968,6 +785,7 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
         CASE_PUT_LABEL(label, pl, Int32);
         CASE_PUT_LABEL(label, pl, Int64);
         default:
+            NPY_END_THREADS;
             PyErr_SetString(PyExc_RuntimeError, "data type not supported");
             goto exit;
         }
@@ -1018,8 +836,9 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
     for (kk = 0; kk < ssize; kk++)
         if (ps[kk] && kk != (ssize / 2))
             ++nneigh;
-    nstrides = (npy_intp*)malloc(nneigh * sizeof(npy_intp));
-    if (!nstrides) {
+    nstrides = malloc(nneigh * sizeof(npy_intp));
+    if (NI_UNLIKELY(!nstrides)) {
+        NPY_END_THREADS;
         PyErr_NoMemory();
         goto exit;
     }
@@ -1079,7 +898,7 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
                     if (!(p->done)) {
                         /* If the neighbor was not processed yet: */
                         int max, pval, vval, wvp, pcost, label, p_idx, v_idx;
-                        switch(input->descr->type_num) {
+                        switch(NI_NormalizeType(input->descr->type_num)) {
                         CASE_WINDEX1(v_index, p_index, strides, input->strides,
                                                  input->nd, i_contiguous, p_idx, v_idx, pi,
                                                  vval, pval, UInt8);
@@ -1087,6 +906,7 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
                                                  input->nd, i_contiguous, p_idx, v_idx, pi,
                                                  vval, pval, UInt16);
                         default:
+                            NPY_END_THREADS;
                             PyErr_SetString(PyExc_RuntimeError,
                                                             "data type not supported");
                             goto exit;
@@ -1104,7 +924,7 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
                                  adapt the cost and the label of the neighbor: */
                             int idx;
                             p->cost = max;
-                            switch(output->descr->type_num) {
+                            switch(NI_NormalizeType(output->descr->type_num)) {
                             CASE_WINDEX2(v_index, strides, output->strides, input->nd,
                                                      idx, o_contiguous, label, pl, UInt8);
                             CASE_WINDEX2(v_index, strides, output->strides, input->nd,
@@ -1124,11 +944,12 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
                             CASE_WINDEX2(v_index, strides, output->strides, input->nd,
                                                      idx, o_contiguous, label, pl, Int64);
                             default:
+                                NPY_END_THREADS;
                                 PyErr_SetString(PyExc_RuntimeError,
                                                                 "data type not supported");
                                 goto exit;
                             }
-                            switch(output->descr->type_num) {
+                            switch(NI_NormalizeType(output->descr->type_num)) {
                             CASE_WINDEX3(p_index, strides, output->strides, input->nd,
                                                      idx, o_contiguous, label, pl, UInt8);
                             CASE_WINDEX3(p_index, strides, output->strides, input->nd,
@@ -1148,6 +969,7 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
                             CASE_WINDEX3(p_index, strides, output->strides, input->nd,
                                                      idx, o_contiguous, label, pl, Int64);
                             default:
+                                NPY_END_THREADS;
                                 PyErr_SetString(PyExc_RuntimeError,
                                                                 "data type not supported");
                                 goto exit;
@@ -1189,13 +1011,10 @@ int NI_WatershedIFT(PyArrayObject* input, PyArrayObject* markers,
         }
     }
  exit:
-    if (temp)
-        free(temp);
-    if (first)
-        free(first);
-    if (last)
-        free(last);
-    if (nstrides)
-        free(nstrides);
+    NPY_END_THREADS;
+    free(temp);
+    free(first);
+    free(last);
+    free(nstrides);
     return PyErr_Occurred() ? 0 : 1;
 }
