@@ -313,13 +313,6 @@ dot_product(const double *u, const double *v, npy_intp n)
     return s;
 }
 
-static NPY_INLINE double
-cosine_distance(const double *u, const double *v, npy_intp n,
-                double nu, double nv)
-{
-    return 1.0 - (dot_product(u, v, n) / (nu * nv));
-}
-
 static NPY_INLINE double
 seuclidean_distance(const double *var, const double *u, const double *v,
                     npy_intp n)
@@ -426,7 +419,7 @@ pdist_cosine(const double *X, double *dm, npy_intp m, npy_intp n,
         for (j = i + 1; j < m; j++, dm++) {
             u = X + (n * i);
             v = X + (n * j);
-            *dm = cosine_distance(u, v, n, norms[i], norms[j]);
+            *dm = 1. - dot_product(u, v, n) / (norms[i] * norms[j]);
         }
     }
 }
@@ -481,13 +474,15 @@ pdist_weighted_minkowski(const double *X, double *dm, npy_intp m, npy_intp n,
 static NPY_INLINE void
 dist_to_squareform_from_vector(double *M, const double *v, npy_intp n)
 {
-    double *it;
+    double *it, *it2;
     npy_intp i, j;
 
     for (i = 0; i < n - 1; i++) {
-        it = M + (i * n) + i + 1;
-        for (j = i + 1; j < n; j++, it++, v++) {
+        it  = M + (i * n) + i + 1;
+        it2 = M + (i+1)*n + i;
+        for (j = i + 1; j < n; j++, it++, it2 += n, v++) {
             *it = *v;
+            *it2 = *v;
         }
     }
 }
@@ -530,23 +525,6 @@ cdist_mahalanobis(const double *XA, const double *XB,
     dimbuf2 = 0;
 }
 
-static NPY_INLINE void
-cdist_cosine(const double *XA, const double *XB, double *dm,
-             npy_intp mA, npy_intp mB, npy_intp n,
-             const double *normsA, const double *normsB)
-{
-    const double *u, *v;
-    npy_intp i, j;
-
-    for (i = 0; i < mA; i++) {
-        for (j = 0; j < mB; j++, dm++) {
-            u = XA + (n * i);
-            v = XB + (n * j);
-            *dm = cosine_distance(u, v, n, normsA[i], normsB[j]);
-        }
-    }
-}
-
 static NPY_INLINE void
 cdist_seuclidean(const double *XA, const double *XB, const double *var,
                  double *dm, npy_intp mA, npy_intp mB, npy_intp n)
