@@ -32,6 +32,17 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#if !defined(__clang__) && defined(__GNUC__) && defined(__GNUC_MINOR__)
+#if __GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 4)
+/* enable auto-vectorizer */
+#pragma GCC optimize("tree-vectorize")
+/* float associativity required to vectorize reductions */
+#pragma GCC optimize("unsafe-math-optimizations")
+/* maybe 5% gain, manual unrolling with more accumulators would be better */
+#pragma GCC optimize("unroll-loops")
+#endif
+#endif
+
 #include <math.h>
 #include <stdlib.h>
 #include <Python.h>
@@ -53,6 +64,7 @@
             return NULL;                                                \
         }                                                               \
         else {                                                          \
+            NPY_BEGIN_ALLOW_THREADS;                                    \
             XA = (const type *)XA_->data;                               \
             XB = (const type *)XB_->data;                               \
             dm = (double *)dm_->data;                                   \
@@ -60,6 +72,7 @@
             mB = XB_->dimensions[0];                                    \
             n = XA_->dimensions[1];                                     \
             cdist_ ## name ## _ ## type(XA, XB, dm, mA, mB, n);         \
+            NPY_END_ALLOW_THREADS;                                      \
         }                                                               \
         return Py_BuildValue("d", 0.);                                  \
     }
@@ -101,6 +114,8 @@ static PyObject *cdist_mahalanobis_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_THREADS_DEF;
+    NPY_BEGIN_THREADS;
     XA = (const double*)XA_->data;
     XB = (const double*)XB_->data;
     covinv = (const double*)covinv_->data;
@@ -111,37 +126,12 @@ static PyObject *cdist_mahalanobis_wrap(PyObject *self, PyObject *args) {
 
     dimbuf = mahalanobis_dimbuf(n);
     if (!dimbuf) {
+      NPY_END_THREADS;
       return NULL;
     }
     cdist_mahalanobis(XA, XB, covinv, dimbuf, dm, mA, mB, n);
     free(dimbuf);
-  }
-  return Py_BuildValue("d", 0.0);
-}
-
-static PyObject *cdist_cosine_wrap(PyObject *self, PyObject *args) {
-  PyArrayObject *XA_, *XB_, *dm_, *normsA_, *normsB_;
-  int mA, mB, n;
-  double *dm;
-  const double *XA, *XB, *normsA, *normsB;
-  if (!PyArg_ParseTuple(args, "O!O!O!O!O!",
-			&PyArray_Type, &XA_, &PyArray_Type, &XB_, 
-			&PyArray_Type, &dm_,
-			&PyArray_Type, &normsA_,
-			&PyArray_Type, &normsB_)) {
-    return 0;
-  }
-  else {
-    XA = (const double*)XA_->data;
-    XB = (const double*)XB_->data;
-    dm = (double*)dm_->data;
-    normsA = (const double*)normsA_->data;
-    normsB = (const double*)normsB_->data;
-    mA = XA_->dimensions[0];
-    mB = XB_->dimensions[0];
-    n = XA_->dimensions[1];
-
-    cdist_cosine(XA, XB, dm, mA, mB, n, normsA, normsB);
+    NPY_END_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -158,6 +148,7 @@ static PyObject *cdist_seuclidean_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const double*)XA_->data;
     XB = (const double*)XB_->data;
     dm = (double*)dm_->data;
@@ -167,6 +158,7 @@ static PyObject *cdist_seuclidean_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_seuclidean(XA, XB, var, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -182,6 +174,7 @@ static PyObject *cdist_hamming_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -190,6 +183,7 @@ static PyObject *cdist_hamming_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_hamming_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -205,6 +199,7 @@ static PyObject *cdist_jaccard_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -213,6 +208,7 @@ static PyObject *cdist_jaccard_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_jaccard_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -230,6 +226,7 @@ static PyObject *cdist_minkowski_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const double*)XA_->data;
     XB = (const double*)XB_->data;
     dm = (double*)dm_->data;
@@ -237,6 +234,7 @@ static PyObject *cdist_minkowski_wrap(PyObject *self, PyObject *args) {
     mB = XB_->dimensions[0];
     n = XA_->dimensions[1];
     cdist_minkowski(XA, XB, dm, mA, mB, n, p);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -255,6 +253,7 @@ static PyObject *cdist_weighted_minkowski_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const double*)XA_->data;
     XB = (const double*)XB_->data;
     w = (const double*)w_->data;
@@ -263,6 +262,7 @@ static PyObject *cdist_weighted_minkowski_wrap(PyObject *self, PyObject *args) {
     mB = XB_->dimensions[0];
     n = XA_->dimensions[1];
     cdist_weighted_minkowski(XA, XB, dm, mA, mB, n, p, w);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -278,6 +278,7 @@ static PyObject *cdist_matching_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -286,6 +287,7 @@ static PyObject *cdist_matching_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_matching_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -301,6 +303,7 @@ static PyObject *cdist_dice_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -309,6 +312,7 @@ static PyObject *cdist_dice_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_dice_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -324,6 +328,7 @@ static PyObject *cdist_rogerstanimoto_bool_wrap(PyObject *self, PyObject *args)
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -332,6 +337,7 @@ static PyObject *cdist_rogerstanimoto_bool_wrap(PyObject *self, PyObject *args)
     n = XA_->dimensions[1];
 
     cdist_rogerstanimoto_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -347,6 +353,7 @@ static PyObject *cdist_russellrao_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -355,6 +362,7 @@ static PyObject *cdist_russellrao_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_russellrao_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -370,6 +378,7 @@ static PyObject *cdist_kulsinski_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -378,6 +387,7 @@ static PyObject *cdist_kulsinski_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_kulsinski_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -393,6 +403,7 @@ static PyObject *cdist_sokalmichener_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -401,6 +412,7 @@ static PyObject *cdist_sokalmichener_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_sokalmichener_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -416,6 +428,7 @@ static PyObject *cdist_sokalsneath_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     XA = (const char*)XA_->data;
     XB = (const char*)XB_->data;
     dm = (double*)dm_->data;
@@ -424,6 +437,7 @@ static PyObject *cdist_sokalsneath_bool_wrap(PyObject *self, PyObject *args) {
     n = XA_->dimensions[1];
 
     cdist_sokalsneath_char(XA, XB, dm, mA, mB, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -443,11 +457,13 @@ static PyObject *cdist_sokalsneath_bool_wrap(PyObject *self, PyObject *args) {
             return NULL;                                                \
         }                                                               \
         else {                                                          \
+            NPY_BEGIN_ALLOW_THREADS;                                    \
             X = (const double *)X_->data;                               \
             dm = (double *)dm_->data;                                   \
             m = X_->dimensions[0];                                      \
             n = X_->dimensions[1];                                      \
             pdist_ ## name ## _double(X, dm, m, n);                     \
+            NPY_END_ALLOW_THREADS;                                      \
         }                                                               \
         return Py_BuildValue("d", 0.);                                  \
     }
@@ -475,6 +491,8 @@ static PyObject *pdist_mahalanobis_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_THREADS_DEF;
+    NPY_BEGIN_THREADS;
     X = (const double*)X_->data;
     covinv = (const double*)covinv_->data;
     dm = (double*)dm_->data;
@@ -483,11 +501,13 @@ static PyObject *pdist_mahalanobis_wrap(PyObject *self, PyObject *args) {
 
     dimbuf = mahalanobis_dimbuf(n);
     if (!dimbuf) {
+        NPY_END_THREADS;
         return NULL;
     }
 
     pdist_mahalanobis(X, covinv, dimbuf, dm, m, n);
     free(dimbuf);
+    NPY_END_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -505,6 +525,7 @@ static PyObject *pdist_cosine_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const double*)X_->data;
     dm = (double*)dm_->data;
     norms = (const double*)norms_->data;
@@ -512,6 +533,7 @@ static PyObject *pdist_cosine_wrap(PyObject *self, PyObject *args) {
     n = X_->dimensions[1];
 
     pdist_cosine(X, dm, m, n, norms);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -528,6 +550,7 @@ static PyObject *pdist_seuclidean_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (double*)X_->data;
     dm = (double*)dm_->data;
     var = (double*)var_->data;
@@ -535,6 +558,7 @@ static PyObject *pdist_seuclidean_wrap(PyObject *self, PyObject *args) {
     n = X_->dimensions[1];
 
     pdist_seuclidean(X, var, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -550,12 +574,14 @@ static PyObject *pdist_hamming_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_hamming_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -571,12 +597,14 @@ static PyObject *pdist_jaccard_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_jaccard_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -593,12 +621,14 @@ static PyObject *pdist_minkowski_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (double*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_minkowski(X, dm, m, n, p);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -616,6 +646,7 @@ static PyObject *pdist_weighted_minkowski_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (double*)X_->data;
     dm = (double*)dm_->data;
     w = (double*)w_->data;
@@ -623,6 +654,7 @@ static PyObject *pdist_weighted_minkowski_wrap(PyObject *self, PyObject *args) {
     n = X_->dimensions[1];
 
     pdist_weighted_minkowski(X, dm, m, n, p, w);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -639,12 +671,14 @@ static PyObject *pdist_yule_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_yule_bool_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -660,12 +694,14 @@ static PyObject *pdist_matching_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_matching_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -681,12 +717,14 @@ static PyObject *pdist_dice_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_dice_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -702,12 +740,14 @@ static PyObject *pdist_rogerstanimoto_bool_wrap(PyObject *self, PyObject *args)
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_rogerstanimoto_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -723,12 +763,14 @@ static PyObject *pdist_russellrao_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_russellrao_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -744,12 +786,14 @@ static PyObject *pdist_kulsinski_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_kulsinski_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -765,12 +809,14 @@ static PyObject *pdist_sokalmichener_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_sokalmichener_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -786,12 +832,14 @@ static PyObject *pdist_sokalsneath_bool_wrap(PyObject *self, PyObject *args) {
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     X = (const char*)X_->data;
     dm = (double*)dm_->data;
     m = X_->dimensions[0];
     n = X_->dimensions[1];
 
     pdist_sokalsneath_char(X, dm, m, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("");
 }
@@ -807,10 +855,12 @@ static PyObject *to_squareform_from_vector_wrap(PyObject *self, PyObject *args)
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     M = (double*)M_->data;
     v = (const double*)v_->data;
     n = M_->dimensions[0];
     dist_to_squareform_from_vector(M, v, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -826,10 +876,12 @@ static PyObject *to_vector_from_squareform_wrap(PyObject *self, PyObject *args)
     return 0;
   }
   else {
+    NPY_BEGIN_ALLOW_THREADS;
     M = (const double*)M_->data;
     v = (double*)v_->data;
     n = M_->dimensions[0];
     dist_to_vector_from_squareform(M, v, n);
+    NPY_END_ALLOW_THREADS;
   }
   return Py_BuildValue("d", 0.0);
 }
@@ -840,7 +892,6 @@ static PyMethodDef _distanceWrapMethods[] = {
   {"cdist_canberra_wrap", cdist_canberra_wrap, METH_VARARGS},
   {"cdist_chebyshev_wrap", cdist_chebyshev_wrap, METH_VARARGS},
   {"cdist_city_block_wrap", cdist_city_block_wrap, METH_VARARGS},
-  {"cdist_cosine_wrap", cdist_cosine_wrap, METH_VARARGS},
   {"cdist_dice_bool_wrap", cdist_dice_bool_wrap, METH_VARARGS},
   {"cdist_euclidean_wrap", cdist_euclidean_wrap, METH_VARARGS},
   {"cdist_sqeuclidean_wrap", cdist_sqeuclidean_wrap, METH_VARARGS},
