@@ -4,19 +4,18 @@
 
 from __future__ import division, print_function, absolute_import
 
+import warnings
+
 import numpy as np
-from scipy.lib.six import xrange
+from scipy._lib.six import xrange
 from numpy import (pi, asarray, floor, isscalar, iscomplex, real, imag, sqrt,
-                   where, mgrid, cos, sin, exp, place, issubdtype, extract,
-                   less, vectorize, inexact, nan, zeros, atleast_1d, sinc)
+                   where, mgrid, sin, place, issubdtype, extract,
+                   less, inexact, nan, zeros, atleast_1d, sinc)
 from ._ufuncs import (ellipkm1, mathieu_a, mathieu_b, iv, jv, gamma, psi, zeta,
                       hankel1, hankel2, yv, kv, gammaln, ndtri, errprint, poch,
-                      binom, xlogy)
-from . import _ufuncs
-import types
+                      binom)
 from . import specfun
 from . import orthogonal
-import warnings
 
 __all__ = ['agm', 'ai_zeros', 'assoc_laguerre', 'bei_zeros', 'beip_zeros',
            'ber_zeros', 'bernoulli', 'berp_zeros', 'bessel_diff_formula',
@@ -43,8 +42,8 @@ class SpecialFunctionWarning(Warning):
 warnings.simplefilter("always", category=SpecialFunctionWarning)
 
 
-def diric(x,n):
-    """Returns the periodic sinc function, also called the Dirichlet function
+def diric(x, n):
+    """Return the periodic sinc function, also called the Dirichlet function.
 
     The Dirichlet function is defined as::
 
@@ -63,29 +62,54 @@ def diric(x,n):
     -------
     diric : ndarray
 
+    Examples
+    --------
+    >>> from scipy import special
+    >>> import matplotlib.pyplot as plt
+
+    >>> x = np.linspace(-8*np.pi, 8*np.pi, num=201)
+    >>> plt.figure(figsize=(8,8));
+    >>> for idx, n in enumerate([2,3,4,9]):
+    ...     plt.subplot(2, 2, idx+1)
+    ...     plt.plot(x, special.diric(x, n))
+    ...     plt.title('diric, n={}'.format(n))
+    >>> plt.show()
+
     """
-    x,n = asarray(x), asarray(n)
+    x, n = asarray(x), asarray(n)
     n = asarray(n + (x-x))
     x = asarray(x + (n-n))
     if issubdtype(x.dtype, inexact):
         ytype = x.dtype
     else:
         ytype = float
-    y = zeros(x.shape,ytype)
+    y = zeros(x.shape, ytype)
+
+    # empirical minval for 32, 64 or 128 bit float computations
+    # where sin(x/2) < minval, result is fixed at +1 or -1
+    if np.finfo(ytype).eps < 1e-18:
+        minval = 1e-11
+    elif np.finfo(ytype).eps < 1e-15:
+        minval = 1e-7
+    else:
+        minval = 1e-3
 
     mask1 = (n <= 0) | (n != floor(n))
-    place(y,mask1,nan)
+    place(y, mask1, nan)
 
-    z = asarray(x / 2.0 / pi)
-    mask2 = (1-mask1) & (z == floor(z))
-    zsub = extract(mask2,z)
-    nsub = extract(mask2,n)
-    place(y,mask2,pow(-1,zsub*(nsub-1)))
+    x = x / 2
+    denom = sin(x)
+    mask2 = (1-mask1) & (abs(denom) < minval)
+    xsub = extract(mask2, x)
+    nsub = extract(mask2, n)
+    zsub = xsub / pi
+    place(y, mask2, pow(-1, np.round(zsub)*(nsub-1)))
 
     mask = (1-mask1) & (1-mask2)
-    xsub = extract(mask,x)
-    nsub = extract(mask,n)
-    place(y,mask,sin(nsub*xsub/2.0)/(nsub*sin(xsub/2.0)))
+    xsub = extract(mask, x)
+    nsub = extract(mask, n)
+    dsub = extract(mask, denom)
+    place(y, mask, sin(nsub*xsub)/(nsub*dsub))
     return y
 
 
@@ -491,7 +515,7 @@ def hyp0f1(v, z):
     .. math:: _0F_1(v,z) = \sum_{k=0}^{\inf}\frac{z^k}{(v)_k k!}.
 
     It's also the limit as q -> infinity of ``1F1(q;v;z/q)``, and satisfies
-    the differential equation :math:``f''(z) + vf'(z) = f(z)`.
+    the differential equation :math:`f''(z) + vf'(z) = f(z)`.
     """
     v = atleast_1d(v)
     z = atleast_1d(z)
@@ -688,7 +712,7 @@ def lpmn(m,n,z):
     return p,pd
 
 
-def clpmn(m,n,z,type=3):
+def clpmn(m, n, z, type=3):
     """Associated Legendre function of the first kind, Pmn(z)
 
     Computes the (associated) Legendre function of the first kind
@@ -710,17 +734,17 @@ def clpmn(m,n,z,type=3):
        Legendre function
     z : float or complex
         Input value.
-    type : int
+    type : int, optional
        takes values 2 or 3
-       2: cut on the real axis |x|>1
-       3: cut on the real axis -1<x<1 (default)
+       2: cut on the real axis ``|x| > 1``
+       3: cut on the real axis ``-1 < x < 1`` (default)
 
     Returns
     -------
     Pmn_z : (m+1, n+1) array
-       Values for all orders 0..m and degrees 0..n
+       Values for all orders ``0..m`` and degrees ``0..n``
     Pmn_d_z : (m+1, n+1) array
-       Derivatives for all orders 0..m and degrees 0..n
+       Derivatives for all orders ``0..m`` and degrees ``0..n``
 
     See Also
     --------
@@ -734,7 +758,7 @@ def clpmn(m,n,z,type=3):
     factor with respect to Ferrer's function of the first kind
     (cf. `lpmn`).
 
-    For ``type=2`` a cut at |x|>1 is chosen. Approaching the real values
+    For ``type=2`` a cut at ``|x| > 1`` is chosen. Approaching the real values
     on the interval (-1, 1) in the complex plane yields Ferrer's function
     of the first kind.
 
@@ -1112,7 +1136,7 @@ def ellipk(m):
     ellipkinc : Incomplete elliptic integral of the first kind
     ellipe : Complete elliptic integral of the second kind
     ellipeinc : Incomplete elliptic integral of the second kind
- 
+
 
     """
     return ellipkm1(1 - asarray(m))
@@ -1167,13 +1191,14 @@ def comb(N, k, exact=False, repetition=False):
 
     Examples
     --------
+    >>> from scipy.special import comb
     >>> k = np.array([3, 4])
     >>> n = np.array([10, 10])
-    >>> sc.comb(n, k, exact=False)
+    >>> comb(n, k, exact=False)
     array([ 120.,  210.])
-    >>> sc.comb(10, 3, exact=True)
+    >>> comb(10, 3, exact=True)
     120L
-    >>> sc.comb(10, 3, exact=True, repetition=True)
+    >>> comb(10, 3, exact=True, repetition=True)
     220L
 
     """
@@ -1227,6 +1252,7 @@ def perm(N, k, exact=False):
 
     Examples
     --------
+    >>> from scipy.special import perm
     >>> k = np.array([3, 4])
     >>> n = np.array([10, 10])
     >>> perm(n, k)
@@ -1280,10 +1306,11 @@ def factorial(n,exact=False):
 
     Examples
     --------
+    >>> from scipy.special import factorial
     >>> arr = np.array([3,4,5])
-    >>> sc.factorial(arr, exact=False)
+    >>> factorial(arr, exact=False)
     array([   6.,   24.,  120.])
-    >>> sc.factorial(5, exact=True)
+    >>> factorial(5, exact=True)
     120L
 
     """
@@ -1328,6 +1355,7 @@ def factorial2(n, exact=False):
 
     Examples
     --------
+    >>> from scipy.special import factorial2
     >>> factorial2(7, exact=False)
     array(105.00000000000001)
     >>> factorial2(7, exact=True)
@@ -1357,7 +1385,7 @@ def factorial2(n, exact=False):
         return vals
 
 
-def factorialk(n,k,exact=True):
+def factorialk(n, k, exact=True):
     """
     n(!!...!)  = multifactorial of order k
     k times
@@ -1366,6 +1394,8 @@ def factorialk(n,k,exact=True):
     ----------
     n : int
         Calculate multifactorial. If `n` < 0, the return value is 0.
+    k : int
+        Order of multifactorial.
     exact : bool, optional
         If exact is set to True, calculate the answer exactly using
         integer arithmetic.
@@ -1382,9 +1412,10 @@ def factorialk(n,k,exact=True):
 
     Examples
     --------
-    >>> sc.factorialk(5, 1, exact=True)
+    >>> from scipy.special import factorialk
+    >>> factorialk(5, 1, exact=True)
     120L
-    >>> sc.factorialk(5, 3, exact=True)
+    >>> factorialk(5, 3, exact=True)
     10L
 
     """
